/*
	G. Wiesenekker. ZZZZZZ a chess program.
	Copyright (C) 1993-1994 G. Wiesenekker
	E-mail: wiesenekker@sara.nl

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "zzzzzz.h"
#ifdef HAS_TIMES
#include <sys/times.h>
#endif
local char *FILE_NAME = __FILE__;

#ifdef PVM

/*
See for example
R. Feldmann, B. Monien and P. Mysliwietz, ICCA june 1989, pp 65-72.

Every time a slaves finds a better score, that score has to be broadcast
to its brothers.
If a processor is ready, it has to inform all other processors that it has
become idle.

As this is the first Massively Parallel Version, I have chosen for 
a setup that is simple to debug.

One processor, the 'slave-driver', keeps track of a stack of all free
processors.
A master sends the moves to the slave driver and requests for a slave.
The slave-driver allocates a slave to the master (removes them from the stack).
As requests for slaves arrive in order at the slave-driver, there is no chance
of allocating the same slave to more than one master.
Each time a slave finds a better score, it sends the score to the 
slave-driver. The slave-driver sends the score to its brothers.
When a slave has finished, it tells the slave-driver that it is ready.
The slave-driver then adds the slave to the stack of free processors.

As messages for score updates and processors becoming idle arrive at random
at the master, there is a chance that an update score message is sent to a 
slave which is already working on another problem. 
To avoid this, a family-id is assigned to brothers. The family-id is unique 
and is incremented at each request for slaves, so a slave should always
ignore messages with a family-id LESS THAN its current family-id.

A problem with the setup is of course that communication with the slave-driver
might become a bottle neck.

*/

#define ZZZZZZ_EXEC "zzzzzz"

/*
Sigh
TO DO: get rid of fixed maximum number of pseudo legal moves.
*/

#define NBUF_MAX 256

/*
brother_tids[0] is always the master.
*/

typedef struct
{
	int nbrothers;
	int brother_tids[PVM_BROTHERS_MAX];
	score_t family_score;
	score_t family_beta;
	int family_quit_tag;
	int master_waiting;
} family_t;

int drivertid;
local family_t family[PVM_FAMILY_MAX];

/*
do_cancel cancels a search process because of a time limit
or a beta cutoff.
*/

local void do_cancel(int);
local void do_cancel(int familyid)
{
	family_t *with;
	score_t temp_score = SCORE_INVAL;
	int ibrother;
	int ifamily;
	
	PVMBUG(pvm_initsend(PvmDataDefault))
	PVMBUG(pvm_pkint(&temp_score, 1, 1))
	with = family + (familyid % PVM_FAMILY_MAX);
	for (ibrother = 0; ibrother < with->nbrothers; ibrother++)
	{
		PVMBUG(pvm_send(with->brother_tids[ibrother], with->family_quit_tag))
		
		/*
		also cancel the families of which this brother is the master.
		*/
		
		if (ibrother > 0)
		{
			for (ifamily = 0; ifamily < PVM_FAMILY_MAX; ifamily++)
			{
				if (family[ifamily].brother_tids[0] == with->brother_tids[ibrother])
					do_cancel(ifamily);
			}
		}
	}
}

void do_master(void)
{
	int mytid;
	char *argd[2] = {"driver", NULL};
	char *args[2] = {"slave", NULL};
	int ntids;
	int tids[PVM_PROCS_MAX];
	
	/*
	Enroll in pvm.
	*/
	
	PVMBUG(mytid = pvm_mytid())
	
	/*
	Spawn slave-driver and slaves.
	*/
	
	ntids = pvm_spawn(ZZZZZZ_EXEC, argd, PvmTaskDefault, NULL, 1, &drivertid);
	BUG(ntids != 1)
	ntids = pvm_spawn(ZZZZZZ_EXEC, args, PvmTaskDefault, NULL, PvmProcs, tids);
	BUG(ntids != PvmProcs)
	
	/*
	Send tids of slaves to driver
	*/

	PVMBUG(pvm_initsend(PvmDataDefault))
	PVMBUG(pvm_pkint(&PvmProcs, 1, 1))
	PVMBUG(pvm_pkint(tids, PvmProcs, 1))
	PVMBUG(pvm_send(drivertid, PVM_SLAVE_TAG))
}

void do_driver(void)
{
	int mytid;
	char name[FILENAME_MAX];
	int nslaves;
	int slave_tids[PVM_PROCS_MAX];
	int familyid;
	int nmoves;
	move_t moves[NBUF_MAX];
	int bufid, nbytes, tag, tid;
	int temp_familyid, current_familyid, has_slave;
	score_t temp_score, best_score;
	family_t *with;
	int ibrother, ifamily;
	char fen_string[LINE_MAX];
	int islave;
	score_t alpha, beta;
	int depth;
	double temp_cp, temp_wall, total_cp;
	time_t wall1, wall2;
	double wall;
#ifdef HAS_TIMES
	struct tms before, after;
	clock_t utime, stime, starttime, endtime;
#endif
	
	/*
	Enroll in pvm.
	*/
	
	wall1 = time(NULL);
#ifdef HAS_TIMES
	starttime = times(&before);
#endif
	PVMBUG(mytid = drivertid = pvm_mytid())
	sprintf(name, "driver.%d.out", mytid);
	my_printf("%z", name);

	/*
	The first message the driver receives contains the tids
	of the slaves.
	*/
	
	PVMBUG(pvm_recv(-1, PVM_SLAVE_TAG))
	PVMBUG(pvm_upkint(&nslaves, 1, 1))
	BUG(nslaves >= PVM_PROCS_MAX)
	PVMBUG(pvm_upkint(slave_tids, nslaves, 1))
	
	for (islave = 0; islave < nslaves; islave++)
	{
		PVMBUG(pvm_initsend(PvmDataDefault))
		PVMBUG(pvm_pkint(&islave, 1, 1))
		PVMBUG(pvm_pkint(&drivertid, 1, 1))
		PVMBUG(pvm_send(slave_tids[islave], PVM_NUMBER_TAG))
	}
	
	/*
	The slave driver waits for requests for slaves, 
	updates of scores, and idle messages
	*/
	
	familyid = PVM_FAMILYID_MIN;
	for (ifamily = 0; ifamily < PVM_FAMILY_MAX; ifamily++)
		family[ifamily].nbrothers = 0;
	
	while(TRUE)
	{
		PVMBUG(bufid = pvm_recv(-1, -1))
		PVMBUG(pvm_bufinfo(bufid, &nbytes, &tag, &tid))
		if (tag == PVM_ALLOCATE_TAG)
		{
			/*
			Ignore the request if there are not enough slaves.
			*/
			
			if (nslaves < 1)
			{
				PVMBUG(pvm_initsend(PvmDataDefault))
				has_slave = FALSE;
				PVMBUG(pvm_pkint(&has_slave, 1, 1))
				PVMBUG(pvm_send(tid, PVM_ALLOCATE_TAG))
				continue;
			}

			PVMBUG(pvm_upkint(&temp_familyid, 1, 1))
			PVMBUG(pvm_upkstr(fen_string))
			PVMBUG(pvm_upkint(&depth, 1, 1))
			PVMBUG(pvm_upkint(&nmoves, 1, 1))
			PVMBUG(pvm_upklong(moves, nmoves, 1))
			
			if (temp_familyid != INVALID)
			{
				current_familyid = temp_familyid;
				with = family + (current_familyid % PVM_FAMILY_MAX);
				if (options.print_distrib)
				{
					my_printf("got alloc from master %d, extending family %d\n",
						tid, current_familyid);
				}
			}
			else
			{
				PVMBUG(pvm_upkint(&best_score, 1, 1))
				PVMBUG(pvm_upkint(&alpha, 1, 1))
				PVMBUG(pvm_upkint(&beta, 1, 1))
				
				/*
				The families are stored in a cyclic buffer that is large enough, so
				there should always be a free family.
				*/
				
				islave = 0;
				while(TRUE)
				{
					current_familyid = familyid;
					with = family + (current_familyid % PVM_FAMILY_MAX);
					familyid++;
					if (with->nbrothers == 0) break;
					islave++;
					BUG(islave >= PVM_PROCS_MAX)
				}
				
				/*
				The first slave is the master.
				*/
				
				with->brother_tids[0] = tid;
				with->nbrothers = 1;
				with->family_quit_tag = 1000000000 + current_familyid;
				with->family_score = best_score;
				with->family_beta = beta;
				with->master_waiting = FALSE;
				if (options.print_distrib)
				{
					my_printf("alloc from master %d, new family is %d\n",
						tid, current_familyid);
				}
			}

			BUG(nslaves < 1)
			if (with->nbrothers == PVM_BROTHERS_MAX)
			{
				my_printf("warning: PVM_BROTHERS_MAX brothers have been allocated!\n");
				PVMBUG(pvm_initsend(PvmDataDefault))
				has_slave = FALSE;
				PVMBUG(pvm_pkint(&has_slave, 1, 1))
				PVMBUG(pvm_send(tid, PVM_ALLOCATE_TAG))
				continue;
			}
			
			with->brother_tids[with->nbrothers] = slave_tids[--nslaves];
			
			PVMBUG(pvm_initsend(PvmDataDefault))
			has_slave = TRUE;
			PVMBUG(pvm_pkint(&has_slave, 1, 1))
			if (temp_familyid == INVALID)
			{
				PVMBUG(pvm_pkint(&current_familyid, 1, 1))
				PVMBUG(pvm_pkint(&(with->family_quit_tag), 1, 1))
			}
			PVMBUG(pvm_send(tid, PVM_ALLOCATE_TAG))
			
			PVMBUG(pvm_initsend(PvmDataDefault))
			PVMBUG(pvm_pkint(&current_familyid, 1, 1))
			PVMBUG(pvm_pkint(&(with->family_quit_tag), 1, 1))
			PVMBUG(pvm_pkstr(fen_string))
			PVMBUG(pvm_pkint(&depth, 1, 1))
			PVMBUG(pvm_pkint(&nmoves, 1, 1))
			PVMBUG(pvm_pklong(moves, nmoves, 1))
			PVMBUG(pvm_pkint(&best_score, 1, 1))
			PVMBUG(pvm_pkint(&alpha, 1, 1))
			PVMBUG(pvm_pkint(&beta, 1, 1))
			PVMBUG(pvm_send(with->brother_tids[with->nbrothers], PVM_ALLOCATE_TAG))
			if (options.print_distrib)
			{
				my_printf("added slave %d to family %d\n",
					with->brother_tids[with->nbrothers], current_familyid);
			}
			with->nbrothers++;
		}
		else if (tag == PVM_UPDATE_TAG)
		{
			PVMBUG(pvm_upkint(&temp_familyid, 1, 1))
			BUG(temp_familyid > familyid)
			PVMBUG(pvm_upkint(&temp_score, 1, 1))
			with = family + (temp_familyid % PVM_FAMILY_MAX);
			if (temp_score > with->family_score)
			{
				with->family_score = temp_score;
				
				/*
				Send the better score to the brothers.
				TO DO: use signals for updates.
				*/
				
				PVMBUG(pvm_initsend(PvmDataDefault))
				PVMBUG(pvm_pkint(&temp_score, 1, 1))
				for (ibrother = 0; ibrother < with->nbrothers; ibrother++)
				{
					/*
					Skip self
					*/
					
					if (with->brother_tids[ibrother] == tid) continue;
					PVMBUG(pvm_send(with->brother_tids[ibrother], temp_familyid))
					if (options.print_distrib)
					{
						my_printf("update score %d from %d to %d in family %d\n",
							temp_score, tid, with->brother_tids[ibrother], temp_familyid);
					}
				}
				
				if (with->family_score >= with->family_beta)
					do_cancel(temp_familyid);
			}
		}
		else if (tag == PVM_WAITING_TAG)
		{
			PVMBUG(pvm_upkint(&temp_familyid, 1, 1))
			with = family + (temp_familyid % PVM_FAMILY_MAX);
			BUG(with->brother_tids[0] != tid)
			with->master_waiting = TRUE;
			if (options.print_distrib)
			{
				my_printf("got waiting from %d in family %d\n",
					tid, temp_familyid);
			}
			if (with->nbrothers == 1)
			{
				PVMBUG(pvm_initsend(PvmDataDefault))
				temp_score = SCORE_INVAL;
				PVMBUG(pvm_pkint(&temp_score, 1, 1))
				PVMBUG(pvm_send(with->brother_tids[0], temp_familyid))
				with->nbrothers--;
				if (options.print_distrib)
				{
					my_printf("send %d to family %d\n",
						temp_score, temp_familyid);
				}
			}
		}
		else if (tag == PVM_IDLE_TAG)
		{
			PVMBUG(pvm_upkint(&temp_familyid, 1, 1))
			BUG(temp_familyid > familyid)
			with = family + (temp_familyid % PVM_FAMILY_MAX);
			if (options.print_distrib)
			{
				my_printf("got idle from slave %d in family %d\n",
					tid, temp_familyid);
			}
			PVMBUG(pvm_initsend(PvmDataDefault))
			temp_score = SCORE_INVAL;
			PVMBUG(pvm_pkint(&temp_score, 1, 1))
			PVMBUG(pvm_send(tid, temp_familyid))
			
			/*
			Remove slave from family.
			*/
			
			for (ibrother = 1; ibrother < with->nbrothers; ibrother++)
				if (with->brother_tids[ibrother] == tid) break;
			BUG(ibrother >= with->nbrothers)
			with->brother_tids[ibrother] = with->brother_tids[with->nbrothers - 1];
			with->nbrothers--;
			BUG(with->nbrothers < 0)
			
			/*
			Make slave free.
			*/
			
			BUG(nslaves >= PVM_PROCS_MAX)
			slave_tids[nslaves++] = tid;
			
			/*
			Inform master if all slaves are gone.
			*/
			
			if (with->master_waiting)
			{
				if (with->nbrothers == 1)
				{
					PVMBUG(pvm_initsend(PvmDataDefault))
					temp_score = SCORE_INVAL;
					PVMBUG(pvm_pkint(&temp_score, 1, 1))
					PVMBUG(pvm_send(with->brother_tids[0], temp_familyid))
					with->nbrothers--;
				}
			}
		}
		else if (tag == PVM_QUIT_TAG)
		{
			total_cp = 0;
			for (islave = 0; islave < nslaves; islave++)
			{
				PVMBUG(pvm_initsend(PvmDataDefault))
				PVMBUG(pvm_pkint(&islave, 1, 1))
				PVMBUG(pvm_send(slave_tids[islave], PVM_QUIT_TAG))
				PVMBUG(pvm_recv(slave_tids[islave], PVM_QUIT_TAG))
				PVMBUG(pvm_upkdouble(&temp_cp, 1, 1))
				PVMBUG(pvm_upkdouble(&temp_wall, 1, 1))
				my_printf("%d %.2f %.2f\n", islave, temp_cp, temp_wall);
				total_cp += temp_cp;
			}
			wall2 = time(NULL);
			wall = difftime(wall2, wall1);
			my_printf("%.2f %.2f\n", total_cp, wall);
#ifdef HAS_TIMES
			endtime = times(&after);
			utime = after.tms_utime - before.tms_utime;
			stime = after.tms_stime - before.tms_stime;

			my_printf("user   cpu %.2f\n", (double) utime / CLK_TCK);
			my_printf("system cpu %.2f\n", (double) stime / CLK_TCK);
			my_printf("wall       %.2f\n", (double) (endtime - starttime) / CLK_TCK);
#endif
			PVMBUG(pvm_exit())
			exit(EXIT_SUCCESS);
		}
		else
			FATAL_ERROR("tag error")
	}
}
#endif
